@@ -64,6 +64,12 @@ enum {
 	DEFAULT_SIZE_OVERHEAD = 1024
 };
 
+/* When the total copy size is over this limit, we try to
+ * forget the data cached for the files, in order to
+ * lower VM pressure
+ */
+#define DROP_CACHE_SIZE_LIMIT (20*1024*1024)
+
 /* in asynch mode the progress callback does a context switch every time
  * it gets called. We'll only call it every now and then to not loose a
  * lot of performance
@@ -1201,6 +1207,7 @@ copy_file_data (GnomeVFSHandle *target_handle,
 	gpointer buffer;
 	const char *write_buffer;
 	GnomeVFSFileSize total_bytes_read;
+	gboolean forget_cache;
 
 	*skip = FALSE;
 
@@ -1210,6 +1217,9 @@ copy_file_data (GnomeVFSHandle *target_handle,
 
 	buffer = g_malloc (block_size);
 	total_bytes_read = 0;
+
+	/* Enable streaming if the total size is large */
+	forget_cache = progress->progress_info->bytes_total >= DROP_CACHE_SIZE_LIMIT;
 	
 	do {
 		GnomeVFSFileSize bytes_read;
@@ -1227,6 +1237,10 @@ copy_file_data (GnomeVFSHandle *target_handle,
 
 			result = gnome_vfs_read (source_handle, buffer,
 						 block_size, &bytes_read);
+			if (forget_cache) {
+				gnome_vfs_forget_cache (source_handle,
+							total_bytes_read, bytes_read);
+			}
 			if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF)
 				retry = handle_error (&result, progress,
 						      error_mode, skip);
@@ -1236,7 +1250,7 @@ copy_file_data (GnomeVFSHandle *target_handle,
 			break;
 		}
 
-		total_bytes_read += total_bytes_read;
+		total_bytes_read += bytes_read;
 		bytes_to_write = bytes_read;
 
 		progress->progress_info->phase = GNOME_VFS_XFER_PHASE_WRITETARGET;
@@ -1256,6 +1270,11 @@ copy_file_data (GnomeVFSHandle *target_handle,
 			bytes_to_write -= bytes_written;
 			write_buffer += bytes_written;
 		} while ((result == GNOME_VFS_OK && bytes_to_write > 0) || retry);
+		
+		if (forget_cache && bytes_to_write == 0) {
+			gnome_vfs_forget_cache (target_handle,
+						total_bytes_read - bytes_read, bytes_read);
+		}
 
 		progress->progress_info->phase = GNOME_VFS_XFER_PHASE_COPYING;
 
@@ -1301,7 +1320,7 @@ xfer_open_source (GnomeVFSHandle **source_handle,
 {
 	GnomeVFSResult result;
 	gboolean retry;
-
+	
 	*skip = FALSE;
 	do {
 		retry = FALSE;
@@ -1367,19 +1386,41 @@ copy_symlink (GnomeVFSURI *source_uri,
 	      GnomeVFSURI *target_uri,
 	      const char *link_name,
 	      GnomeVFSXferErrorMode *error_mode,
+	      GnomeVFSXferOverwriteMode *overwrite_mode,
 	      GnomeVFSProgressCallbackState *progress,
 	      gboolean *skip)
 {
 	GnomeVFSResult result;
 	gboolean retry;
+	gboolean tried_remove, remove;
+
+	tried_remove = FALSE;
 
 	*skip = FALSE;
 	do {
 		retry = FALSE;
-		
+
 		result = gnome_vfs_create_symbolic_link (target_uri, link_name);
 
-		if (result != GNOME_VFS_OK) {
+		if (result == GNOME_VFS_ERROR_FILE_EXISTS) {
+			remove = FALSE;
+			
+			if (*overwrite_mode == GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE &&
+			    !tried_remove) {
+				remove = TRUE;
+				tried_remove = TRUE;
+			} else {
+				retry = handle_overwrite (&result,
+							  progress,
+							  error_mode,
+							  overwrite_mode,
+							  &remove,
+							  skip);
+			}
+			if (remove) {
+				gnome_vfs_unlink_from_uri (target_uri);
+			}
+		} else if (result != GNOME_VFS_OK) {
 			retry = handle_error (&result, progress, error_mode, skip);
 		} else if (result == GNOME_VFS_OK &&
 			   call_progress_with_uris_often (progress, source_uri,
@@ -1505,9 +1546,11 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 	GnomeVFSResult result;
 	GnomeVFSDirectoryHandle *source_directory_handle;
 	GnomeVFSDirectoryHandle *dest_directory_handle;
+	GnomeVFSFileInfo *target_dir_info;
 
 	source_directory_handle = NULL;
 	dest_directory_handle = NULL;
+	target_dir_info = NULL;
 	
 	result = gnome_vfs_directory_open_from_uri (&source_directory_handle, source_dir_uri, 
 						    GNOME_VFS_FILE_INFO_DEFAULT);
@@ -1541,6 +1584,14 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 		return result;
 	}
 
+	target_dir_info = gnome_vfs_file_info_new ();
+	result = gnome_vfs_get_file_info_uri (target_dir_uri, target_dir_info,
+                                        GNOME_VFS_FILE_INFO_DEFAULT);
+	if (result != GNOME_VFS_OK) {
+		gnome_vfs_file_info_unref (target_dir_info);
+		target_dir_info = NULL;
+	}
+
 	if (call_progress_with_uris_often (progress, source_dir_uri, target_dir_uri, 
 					   GNOME_VFS_XFER_PHASE_OPENTARGET) != 0) {
 
@@ -1567,6 +1618,9 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 				gnome_vfs_file_info_unref (info);	
 				break;
 			}
+			if (target_dir_info && GNOME_VFS_FILE_INFO_SGID(target_dir_info)) {
+				info->gid = target_dir_info->gid;
+			}
 			
 			/* Skip "." and "..".  */
 			if (strcmp (info->name, ".") != 0 && strcmp (info->name, "..") != 0) {
@@ -1596,7 +1650,7 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 						gnome_vfs_file_info_unref (symlink_target_info);
 					} else {
 						result = copy_symlink (source_uri, dest_uri, info->symlink_name,
-								       error_mode, progress, &skip_child);
+								       error_mode, overwrite_mode, progress, &skip_child);
 					}
 				}
 				/* We don't want to overwrite a previous skip with FALSE, so we only
@@ -1628,20 +1682,29 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 	gnome_vfs_directory_close (source_directory_handle);
 
 	if (result == GNOME_VFS_OK) {
+		GnomeVFSFileInfo *info;
+		info = gnome_vfs_file_info_new ();
 		/* FIXME the modules should ignore setting of permissions if
 		 * "valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS" is clear
 		 * for now, make sure permissions aren't set to 000
 		 */
-
+		gnome_vfs_file_info_copy (source_file_info, info);
+		if (target_dir_info && GNOME_VFS_FILE_INFO_SGID(target_dir_info)) {
+			info->gid = target_dir_info->gid;
+		}
 		if ((source_file_info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS) != 0) {
 			/* Call this separately from the time one, since one of them may fail,
 			   making the other not run. */
-			gnome_vfs_set_file_info_uri (target_dir_uri, source_file_info, 
+			gnome_vfs_set_file_info_uri (target_dir_uri, info, 
 						     GNOME_VFS_SET_FILE_INFO_OWNER | GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
 		}
 		
 		/* Call this last so nothing else changes the times */
-		gnome_vfs_set_file_info_uri (target_dir_uri, source_file_info, GNOME_VFS_SET_FILE_INFO_TIME);
+		gnome_vfs_set_file_info_uri (target_dir_uri, info, GNOME_VFS_SET_FILE_INFO_TIME);
+		gnome_vfs_file_info_unref (info);
+	}
+	if (target_dir_info) {
+		gnome_vfs_file_info_unref (target_dir_info);
 	}
 
 	return result;
@@ -1691,6 +1754,17 @@ copy_items (const GList *source_uri_list,
 			((GnomeVFSURI *)target_item->data);
 
 		if (result == GNOME_VFS_OK) {
+			GnomeVFSFileInfo *target_dir_info;
+			/* get target_dir URI and file info to take care of SGID */
+			target_dir_info = gnome_vfs_file_info_new ();
+			result = gnome_vfs_get_file_info_uri (target_dir_uri, target_dir_info,
+																						GNOME_VFS_FILE_INFO_DEFAULT);
+			if (result == GNOME_VFS_OK && GNOME_VFS_FILE_INFO_SGID(target_dir_info)) {
+				info->gid = target_dir_info->gid;
+			}
+			gnome_vfs_file_info_unref (target_dir_info);
+			result = GNOME_VFS_OK;
+
 			/* optionally keep trying until we hit a unique target name */
 			for (count = 1; ; count++) {
 				GnomeVFSXferOverwriteMode overwrite_mode_abort;
@@ -1724,7 +1798,8 @@ copy_items (const GList *source_uri_list,
 								 progress, &skip);
                                 } else if (info->type == GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK) {
 					result = copy_symlink (source_uri, target_uri, info->symlink_name,
-							       error_mode, progress, &skip);
+							       error_mode, &overwrite_mode_abort,
+							       progress, &skip);
                                 }
 				/* just ignore all the other special file system objects here */
 
@@ -2074,6 +2149,7 @@ gnome_vfs_xfer_delete_items_common (const GList *source_uri_list,
 						      GNOME_VFS_FILE_INFO_DEFAULT);
 
 		if (result != GNOME_VFS_OK) {
+			gnome_vfs_file_info_unref (info);
 			break;
 		}
 
@@ -2085,6 +2161,8 @@ gnome_vfs_xfer_delete_items_common (const GList *source_uri_list,
 			result = remove_file (uri, progress, xfer_options, &error_mode,
 					      &skip);
 		}
+
+		gnome_vfs_file_info_unref (info);
 	}
 
 	return result;
@@ -2176,6 +2254,8 @@ gnome_vfs_new_directory_with_unique_name (const GnomeVFSURI *target_dir_uri,
 		gnome_vfs_uri_unref (target_uri);
 	}
 
+	progress->progress_info->vfs_status = result;
+	
 	call_progress_uri (progress, NULL, target_uri,
 		GNOME_VFS_XFER_PHASE_OPENTARGET);
 
@@ -2189,38 +2269,34 @@ gnome_vfs_new_directory_with_unique_name (const GnomeVFSURI *target_dir_uri,
 }
 
 static GnomeVFSResult
-gnome_vfs_destination_is_writable (const GnomeVFSURI *uri)
+gnome_vfs_destination_is_writable (GnomeVFSURI *uri)
 {
-	GnomeVFSURI *test_uri;
 	GnomeVFSResult result;
-	GnomeVFSHandle *handle;
-
-	if (!gnome_vfs_uri_is_local (uri)) {
-		/* if destination is not local, do not test it for writability, just
-		 * assume it's writable
-		 */
-		return GNOME_VFS_OK;
-	}
-
-	/* test writability by creating and erasing a temporary file */
-	test_uri = gnome_vfs_uri_append_file_name (uri, ".vfs-write.tmp");
-	result = gnome_vfs_create_uri (&handle, test_uri, GNOME_VFS_OPEN_WRITE, TRUE, 0600);
-
-	if (result == GNOME_VFS_OK) {
-		gnome_vfs_close (handle);
-	}
+	GnomeVFSFileInfo *info;
 	
-	if (result == GNOME_VFS_OK || result == GNOME_VFS_ERROR_FILE_EXISTS) {
-		gnome_vfs_unlink_from_uri (test_uri);
-		result = GNOME_VFS_OK;
-	}
+	info   = gnome_vfs_file_info_new ();	    
+	result = gnome_vfs_get_file_info_uri (uri, info,
+					      GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS);
 	
-	/* some methods only allow certain filenames (e.g. .desktop files) */
-	if (result == GNOME_VFS_ERROR_INVALID_URI) {
-		result = GNOME_VFS_OK;
+		
+	if (result != GNOME_VFS_OK) {
+		gnome_vfs_file_info_unref (info);
+		return result;
 	}
 
-	gnome_vfs_uri_unref (test_uri);
+	/* Default to GNOME_VFS_OK so we return that if we don't have support
+	   for GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS */
+	result = GNOME_VFS_OK;
+	
+	if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_ACCESS) {
+		
+		if (! (info->permissions & GNOME_VFS_PERM_ACCESS_WRITABLE &&
+		       info->permissions & GNOME_VFS_PERM_ACCESS_EXECUTABLE)) {
+			result = GNOME_VFS_ERROR_ACCESS_DENIED;
+		}
+	}
+
+	gnome_vfs_file_info_unref (info);
 	return result;
 }
 
